# Third Party imports
import os
import subprocess
import statistics
import logging
from pathlib import Path

# Base imports
from jcalc.core.pdb import JCalcPdb
from jcalc.settings import GROMACS_VERSION


class JCalcMd:
    """
    """

    def __init__(self, xtc, tpr, suffix, skip, j_input):

        self.wkdir = Path.cwd()
        self.xtc = xtc
        self.tpr = tpr
        self.suffix = suffix
        self.skip = skip
        self.j_input = self.wkdir.joinpath(j_input)
        self.frames_dir = self.wkdir.joinpath(f"frames{self.suffix}")

    def create_frames(self):
        """ Description:
              Given a JCalcMd struct, separate its XTC file in PDB files,
              skipping n frames chosen (self.skip)

            Usage:
              JCalcMd.create_frames()
        """
        frames_dir = self.frames_dir
        if frames_dir.exists():
            logging.error(f"Dir {str(frames_dir.resolve())} exists, \
please rename it or remove it")
        else:
            frames_dir.mkdir()
        logging.info(f"Creating frames, path: {str(frames_dir.resolve())}")

        subprocess.call(f"echo non-Water non-Water | \
                          {GROMACS_VERSION} trjconv -s {self.tpr} \
                          -f {self.xtc} -sep -skip {self.skip} \
                          -o {str(frames_dir.resolve())}/frame_.pdb -pbc mol \
                          -center",
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.STDOUT,
                        shell=True
                        )
        frames = os.listdir(str(frames_dir.resolve()))
        frames = sorted(frames,
                        key=lambda x: int(x.split("_")[1].split(".")[0])
                        )
        self.frames = frames

    def rename_hydro(self, pdb):
        """ Description:
              Given an PDB file, rename all added hydrogen by obabel so
              BioPDBParser can accept its format

            Usage:
              JCalcMd.rename_hydro()

            Parameters:
              pdb:
                string, pdb filename
        """

        old_h = []
        new_file_lines = []
        with open(pdb, "r") as file:
            for line in file:
                cur_line = line.split()
                try:
                    if "H" in cur_line[2] and \
                     (cur_line[0] == "ATOM" or cur_line[0] == "HETATM"):
                        if cur_line[2] != "H" and cur_line[2] not in old_h:
                            old_h.append(cur_line[2])

                    new_file_lines.append(line)
                except Exception:
                    continue

        # Now, we have all hidrogens names, and we can add new ones without
        # messing with original hydrogens
        counter = 1
        with open(pdb, "w") as file:
            for line in new_file_lines:
                cur_line = line.split()
                try:
                    if cur_line[2] == "H" and \
                     (cur_line[0] == "ATOM" or cur_line[0] == "HETATM"):
                        h_name = f"H{counter}"
                        while h_name in old_h:
                            counter += 1
                            h_name = f"H{counter}"
                        line = line[:12] + f"{h_name:^4s}" + f" {line[17:]}"
                        old_h.append(h_name)
                        file.write(line)
                    else:
                        file.write(line)
                except Exception as error:
                    logging.warning(f"Error {error} at renaming hydrogen")
                    file.write(line)

    def add_hydrogen(self):
        """ Description:
              Given a JCalcMd struct, add hydrogens to all frames from
              the simulation

            Usage:
              JCalcMd.add_hydrogen()
        """
        logging.info("Adding hydrogen to frames (GROMOS non-polar hydrogen)")
        frames_hydro = self.frames_dir.joinpath("hydro")
        frames_hydro.mkdir()

        cmd_add = (
            f"cd {str(frames_hydro.resolve())} && "
            f"obabel -ipdb ../*.pdb -opdb -h -m && "
            "mv * ../ && "
            "cd ../ && "
            "rm -r hydro"
        )
        subprocess.call(
            cmd_add,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.STDOUT,
            shell=True
        )

        frames = os.listdir(str(self.frames_dir.resolve()))
        frames = sorted(
            frames,
            key=lambda x: int(x.split("_")[1].split(".")[0])
        )
        self.frames = frames

        for pdb in frames:
            self.rename_hydro(f"{str(self.frames_dir.resolve())}/{pdb}")

    def calc_md_j(self):
        """ Description:

            Usage:
        """

        logging.info("Calculating J values from MD")
        frames_dir = self.frames_dir
        all_j_values = {}
        n_frames = 0

        for pdb in self.frames:
            j_struct = JCalcPdb(pdb=f"{str(frames_dir.resolve())}/{pdb}",
                                j_input=self.j_input
                                )
            j_struct.get_atoms_vector()
            j_struct.create_j_dict()
            j_struct.calc_all_j()
            all_j_values[pdb] = j_struct
            n_frames += 1

        self.n_frames = n_frames
        self.all_j_values = all_j_values

        # Get J names
        j_names = []
        first_frame = list(self.all_j_values.keys())[0]
        for j in all_j_values[first_frame].j_values:
            j_names.append(j)
        self.j_names = j_names

    def calc_statistics(self):
        """ Description:
              Given a JCalcMd struct, calculate all statistics from J values
              calculated through Molecular Dynamics

            Usage:
              JCalcMd.calc_statistics()
        """

        logging.info("Calculating Statistics from J values")
        statistics_dict = {}

        for j in self.j_names:
            statistics_dict[j] = []

        for pdb in self.all_j_values:
            for j_name, j_value in self.all_j_values[pdb].j_values.items():
                statistics_dict[j_name].append(j_value)

        # Now, calc statistics
        # Mean calc
        mean_results = {}
        for j in statistics_dict:
            mean_results[j] = statistics.mean(statistics_dict[j])
        self.mean_results = mean_results

        # Stdev calc
        stdev_results = {}
        for j in statistics_dict:
            stdev_results[j] = statistics.stdev(statistics_dict[j])

        self.stdev_results = stdev_results

    def write_statistics(self, out_name):
        """ Description:
              Given an JCalcMd and a output filename, returns statistical
              results from coupling constant (J) values

            Usage:
              JCalcMd.write_statistics("statistical_results.txt")

            Parameters:
              out_name:
                string, statitics output file name
        """

        with open(str(out_name), "w") as out:
            for j, mean_value in self.mean_results.items():
                out.write(f"{j}_mean:\t{round(mean_value,2)}\n")
                out.write(f"{j}_stdev:\t{round(self.stdev_results[j],2)}\n")

    def write_j_values(self):
        """ Description:
              Given a JCalcMd, writes all J values for all frames in the
              Molecular Dynamics

            Usage:
            JCalcMd.write_j_values()
        """

        for j in self.j_names:
            out_file = self.wkdir.joinpath(f"{j}_values.tsv")
            logging.info(f"Writing J {j} values through MD, path: \
{str(out_file)}")
            with open(str(out_file), "w") as j_file:
                for pdb in self.all_j_values:
                    j_value = self.all_j_values[pdb].j_values[j]
                    j_file.write(f"{pdb}\t{round(j_value,2)}\n")

    def write_results(self, stats_filename):

        """ Description:
              Given a JCalcMd struct, writes every output possible, being:
              statistics output;
              J values through frames;

            Usage:
              JCalcMd.write_results(stats_filename="statistical_results.txt")

            Parameters:
              stats_filename:
                string, output statistics file name
        """

        # Write statistics results
        stats_file = self.wkdir.joinpath(f"{stats_filename}")
        self.write_statistics(out_name=stats_file)
        logging.info(f"Writing J statistics resuls, path: \
{str(stats_file.resolve())}")

        # Write J values results through Molecular Dynamics
        self.write_j_values()



import logging


def log_config(log_file):
    """ Description:
          Function to start log, writing inside LOGS

        Usage:
          log_config("/home/ubuntu/LOGS/jcalc.log")

        Parameters:
          log_file:
            string, LOGS directory from jcalc
    """
    # Log configurations
    logging.basicConfig(
        filename=log_file, filemode="w",
        format="%(asctime)s - %(message)s",
        datefmt='%m/%d/%Y %H:%M:%S',
        level=logging.INFO
    )



# Third Party imports
import math
import logging
from pathlib import Path
from Bio.PDB import (
    PDBParser,
    Selection,
    NeighborSearch,
    calc_dihedral
)


# Base imports
from jcalc.settings import HUGGINS_ELECTRO


class JCalcPdb:
    """ Class to store vicinal coupling constant (3JH,H) for a given PDB
        structure. Receives as input PDB filename and J input file with
        chosen JH,H
    """

    def __init__(self, pdb, j_input):

        parser = PDBParser(QUIET=True)
        self.wkdir = Path.cwd()
        self.pdb = pdb.replace(".pdb", "")
        self.struct = parser.get_structure(pdb, pdb)
        self.atom_list = Selection.unfold_entities(self.struct, 'A')
        self.j_input = j_input
        self.parse_j_list()

    def parse_j_list(self):
        """ Description:
              Given a JCalc struct, create new attribute with all chosen
              Vicinal Coupling Constant to be calculated, being the attribute
              a list where each item is:
              list[0] = first proton name  ("H1")
              list[1] = second proton name ("H2")

            Usage:
                JCalcPdb.parse_j_list()
        """

        j_list = []
        n_j = 0
        with open(str(self.j_input), "r") as file:
            for line in file:
                line = line.split("\t")
                line[-1] = line[-1].replace("\n", "")
                j_list.append(list(map(int, line)))
                n_j += 1

        self.n_j = n_j
        self.j_list = j_list

    def get_atoms_vector(self):
        """ Description:
              Given an JCalcPdb struct, get all atoms vectors and elements
              from PDB struct

            Usage:
              JCalcPdb.get_atoms_vector()
        """

        structure = self.struct
        atom_dict = {}
        # Pegar so o residuo que quero como struct
        for residue in structure.get_residues():
            for atom in residue:
                atom_dict[atom.serial_number] = \
                 [atom.get_vector(), atom.element, atom.get_coord()]

        self.atom_dict = atom_dict

    def search_subs(
        self,
        hx_atom,
        cx_atom,
        cy_atom,
        hy_atom,
        j_atoms,
        chosen_j
    ):
        """ Description:

            Usage:

            Parameters:
        """

        subs_list = []
        # j_atoms = [j[0] for j in j_atoms]
        center = self.atom_dict[cy_atom][2]
        ns = NeighborSearch(self.atom_list)
        neighbors = ns.search(center, 1.7)
        for neigh_atom in neighbors:
            if neigh_atom.serial_number not in j_atoms:
                subs_list.append(neigh_atom)

        for subs in subs_list:
            self.j_dict[chosen_j]["substituents"][subs.serial_number] = {
                "SY": self.atom_dict[subs.serial_number][0],
                "HX": self.atom_dict[hx_atom][0],
                "CX": self.atom_dict[cx_atom][0],
                "CY": self.atom_dict[cy_atom][0],
                "element": self.atom_dict[subs.serial_number][1]
            }

    def create_j_dict(self):
        """ Description:
              Given an JCalcPdb struct, create J dictioary with chosen 3JH,H
              information to calc all J values

            Usage:
              JcalcPdb.create_j_dict()
        """

        j_dict = {}

        for j in self.j_list:
            chosen_j = f"{j[0]},{j[3]}"
            j_dict[chosen_j] = {}
            j_dict[chosen_j]["HX"] = self.atom_dict[j[0]][0]
            j_dict[chosen_j]["CX"] = self.atom_dict[j[1]][0]
            j_dict[chosen_j]["CY"] = self.atom_dict[j[2]][0]
            j_dict[chosen_j]["HY"] = self.atom_dict[j[3]][0]
            j_dict[chosen_j]["dih"] = calc_dihedral(self.atom_dict[j[0]][0],
                                                    self.atom_dict[j[1]][0],
                                                    self.atom_dict[j[2]][0],
                                                    self.atom_dict[j[3]][0]
                                                    )
            j_dict[chosen_j]["dih"] = math.degrees(j_dict[chosen_j]["dih"])
            j_dict[chosen_j]["substituents"] = {}
            self.j_dict = j_dict

            # Search subs for CX and create subs dict
            self.search_subs(hx_atom=j[0],
                             cx_atom=j[1],
                             cy_atom=j[2],
                             hy_atom=j[3],
                             j_atoms=j,
                             chosen_j=chosen_j
                             )
            # Search subs for CY and create subs dict
            self.search_subs(hx_atom=j[3],
                             cx_atom=j[2],
                             cy_atom=j[1],
                             hy_atom=j[0],
                             j_atoms=j,
                             chosen_j=chosen_j
                             )

    def calc_subs_coupling(self, HX, CX, CY, SY, dihedral, element):
        """ Description:
              Given an JCalcPdb struct and atom vectors, calculate
              J pertubation from substitute atoms

            Usage:
              JcalcPdb.calc_subs_coupling(vector_HX, vector_CX, vector_CY,
                                          vector_SY, HX-CX-CY-XY_dih,
                                          substitue_element
                                         )

            Parameters:
              HX:
                Atom coordinates vector, HX from HX-CX-CY-SY fragment
              CX:
                Atom coordinates vector, CX from HX-CX-CY-SY fragment
              CY:
                Atom coordinates vector, CY from HX-CX-CY-SY fragment
              SY:
                Atom coordinates vector, SY from HX-CX-CY-SY fragment
              dihedral:
                float, dihedral from JH,H fragment HX-CX-CY-XY in radians
              element:
                string, atom element from substitute atom (H, O, N)
        """

        huggins_constant = HUGGINS_ELECTRO[element]
        subs_dih = calc_dihedral(HX, CX, CY, SY)

        if subs_dih >= 0:
            return huggins_constant * \
                (0.56 + (-2.32 * (math.cos(math.radians((dihedral * -1) +
                 (17.9 * huggins_constant))) ** 2)))

        else:
            return huggins_constant * \
                (0.56 + (-2.32 * (math.cos(math.radians(dihedral +
                 (17.9 * huggins_constant))) ** 2)))

    def calc_j_h_h(self, chosen_j):
        """ Description:
              Given an JCalcPdb struct and a chosen 3JH,H, calculate it's J
              value and return it

            Usage:
              JCalcPdb.calc_j_h_h(chosen_j="H1,H2")

            Parameters:
              chosen_j:
                string, chosen 3JH,H value from input couplings contants
                example: "H1,H2"
        """

        subs_value = 0
        for subs in self.j_dict[chosen_j]["substituents"].values():
            subs_value += \
                self.calc_subs_coupling(
                    HX=subs["HX"],
                    CX=subs["CX"],
                    CY=subs["CY"],
                    SY=subs["SY"],
                    dihedral=self.j_dict[chosen_j]["dih"],
                    element=subs["element"]
                    )

        dih_radians = math.radians(self.j_dict[chosen_j]["dih"])
        j_value = (13.86 * (math.cos(dih_radians)) ** 2) + \
                  (-0.81 * math.cos(dih_radians)) + subs_value
        return j_value

    def calc_all_j(self):
        """ Description:
              Given an JcalcPdb struct, call all 3JH,H values given as inputs
              from j_input attribute

            Usage:
              JcalcPdb.calc_all_j()
        """

        j_values = {}
        for j in self.j_dict.keys():
            j_values[j] = self.calc_j_h_h(j)

        self.j_values = j_values

    def write_pdb_results(self):
        """ Description:

            Usage:

            Parameters:
        """

        out_file = self.wkdir.joinpath(f"{self.pdb}_J_values.tsv")
        with open(str(out_file), "w") as j_file:
            for j, j_value in self.j_values.items():
                j_value = self.j_values[j]
                j_file.write(f"{j}\t{round(j_value,2)}\n")
        logging.info(f"Output file path: {str(out_file.resolve())}")



# Base imports
from setuptools import setup

# Project imports
from jcalc.settings import JCALC_VERSION

setup(
    name="jcalc",
    version=JCALC_VERSION,
    description="jcalc: Calculate NMR J values from \
        Molecular Dynamics simulations",
    url="https://github.com/Joaodemeirelles/jcalc/",
    author="Joao Luiz de Meirelles",
    author_email="jldemeirelles@gmail.com",
    licence="Academic",
    packages=[
        "jcalc",
        "jcalc.core",
        "jcalc.logger",
    ],
    scripts=["jcalc/jcalc"],
    install_requires=[
        "numpy==1.19.4",
        "statistics==1.0.3.5",
        "biopython==1.78",
        "argparse==1.4.0",
        "pathlib==1.0.1",
        "flake8==3.9.0"
    ],
    zip_safe=False
)
